@@ -15,7 +15,6 @@
  */
 
 
-#include "mongoc-config.h"
 #include "mongoc-error.h"
 #include "mongoc-read-prefs-private.h"
 #include "mongoc-trace-private.h"
@@ -30,6 +29,7 @@ mongoc_read_prefs_new (mongoc_read_mode_t mode)
    read_prefs->mode = mode;
    bson_init (&read_prefs->tags);
    read_prefs->max_staleness_seconds = MONGOC_NO_MAX_STALENESS;
+   bson_init (&read_prefs->hedge);
 
    return read_prefs;
 }
@@ -93,6 +93,8 @@ mongoc_read_prefs_add_tag (mongoc_read_prefs_t *read_prefs, const bson_t *tag)
    } else {
       bson_append_document (&read_prefs->tags, str, -1, &empty);
    }
+
+   bson_destroy (&empty);
 }
 
 
@@ -116,17 +118,43 @@ mongoc_read_prefs_set_max_staleness_seconds (mongoc_read_prefs_t *read_prefs,
 }
 
 
+const bson_t *
+mongoc_read_prefs_get_hedge (const mongoc_read_prefs_t *read_prefs)
+{
+   BSON_ASSERT (read_prefs);
+
+   return &read_prefs->hedge;
+}
+
+
+void
+mongoc_read_prefs_set_hedge (mongoc_read_prefs_t *read_prefs,
+                             const bson_t *hedge)
+{
+   BSON_ASSERT (read_prefs);
+
+   bson_destroy (&read_prefs->hedge);
+
+   if (hedge) {
+      bson_copy_to (hedge, &read_prefs->hedge);
+   } else {
+      bson_init (&read_prefs->hedge);
+   }
+}
+
+
 bool
 mongoc_read_prefs_is_valid (const mongoc_read_prefs_t *read_prefs)
 {
    BSON_ASSERT (read_prefs);
 
    /*
-    * Tags or maxStalenessSeconds are not supported with PRIMARY mode.
+    * Tags, maxStalenessSeconds, and hedge are not supported with PRIMARY mode.
     */
    if (read_prefs->mode == MONGOC_READ_PRIMARY) {
       if (!bson_empty (&read_prefs->tags) ||
-          read_prefs->max_staleness_seconds != MONGOC_NO_MAX_STALENESS) {
+          read_prefs->max_staleness_seconds != MONGOC_NO_MAX_STALENESS ||
+          !bson_empty (&read_prefs->hedge)) {
          return false;
       }
    }
@@ -136,6 +164,7 @@ mongoc_read_prefs_is_valid (const mongoc_read_prefs_t *read_prefs)
       return false;
    }
 
+
    return true;
 }
 
@@ -145,6 +174,7 @@ mongoc_read_prefs_destroy (mongoc_read_prefs_t *read_prefs)
 {
    if (read_prefs) {
       bson_destroy (&read_prefs->tags);
+      bson_destroy (&read_prefs->hedge);
       bson_free (read_prefs);
    }
 }
@@ -157,8 +187,11 @@ mongoc_read_prefs_copy (const mongoc_read_prefs_t *read_prefs)
 
    if (read_prefs) {
       ret = mongoc_read_prefs_new (read_prefs->mode);
+      bson_destroy (&ret->tags);
       bson_copy_to (&read_prefs->tags, &ret->tags);
       ret->max_staleness_seconds = read_prefs->max_staleness_seconds;
+      bson_destroy (&ret->hedge);
+      bson_copy_to (&read_prefs->hedge, &ret->hedge);
    }
 
    return ret;
@@ -192,43 +225,50 @@ static void
 _apply_read_preferences_mongos (
    const mongoc_read_prefs_t *read_prefs,
    const bson_t *query_bson,
-   mongoc_apply_read_prefs_result_t *result /* OUT */)
+   mongoc_assemble_query_result_t *result /* OUT */)
 {
    mongoc_read_mode_t mode;
    const bson_t *tags = NULL;
    bson_t child;
    const char *mode_str;
-   int64_t max_staleness_seconds;
+   int64_t max_staleness_seconds = MONGOC_NO_MAX_STALENESS;
+   const bson_t *hedge = NULL;
 
    mode = mongoc_read_prefs_get_mode (read_prefs);
    if (read_prefs) {
+      max_staleness_seconds =
+         mongoc_read_prefs_get_max_staleness_seconds (read_prefs);
+
       tags = mongoc_read_prefs_get_tags (read_prefs);
+      hedge = mongoc_read_prefs_get_hedge (read_prefs);
    }
 
    /* Server Selection Spec says:
     *
-    * For mode 'primary', drivers MUST NOT set the slaveOK wire protocol flag
+    * For mode 'primary', drivers MUST NOT set the secondaryOk wire protocol flag
     *   and MUST NOT use $readPreference
     *
-    * For mode 'secondary', drivers MUST set the slaveOK wire protocol flag and
+    * For mode 'secondary', drivers MUST set the secondaryOk wire protocol flag and
     *   MUST also use $readPreference
     *
-    * For mode 'primaryPreferred', drivers MUST set the slaveOK wire protocol
+    * For mode 'primaryPreferred', drivers MUST set the secondaryOk wire protocol
     *   flag and MUST also use $readPreference
     *
-    * For mode 'secondaryPreferred', drivers MUST set the slaveOK wire protocol
+    * For mode 'secondaryPreferred', drivers MUST set the secondaryOk wire protocol
     *   flag. If the read preference contains a non-empty tag_sets parameter,
-    *   drivers MUST use $readPreference; otherwise, drivers MUST NOT use
-    *   $readPreference
+    *   maxStalenessSeconds is a positive integer, or the hedge parameter is
+    *   non-empty, drivers MUST use $readPreference; otherwise, drivers MUST NOT
+    *   use $readPreference
     *
-    * For mode 'nearest', drivers MUST set the slaveOK wire protocol flag and
+    * For mode 'nearest', drivers MUST set the secondaryOk wire protocol flag and
     *   MUST also use $readPreference
     */
-   if (mode == MONGOC_READ_SECONDARY_PREFERRED && bson_empty0 (tags)) {
-      result->flags |= MONGOC_QUERY_SLAVE_OK;
+   if (mode == MONGOC_READ_SECONDARY_PREFERRED &&
+       (bson_empty0 (tags) && max_staleness_seconds <= 0 && bson_empty0 (hedge))) {
+      result->flags |= MONGOC_QUERY_SECONDARY_OK;
 
    } else if (mode != MONGOC_READ_PRIMARY) {
-      result->flags |= MONGOC_QUERY_SLAVE_OK;
+      result->flags |= MONGOC_QUERY_SECONDARY_OK;
 
       /* Server Selection Spec: "When any $ modifier is used, including the
        * $readPreference modifier, the query MUST be provided using the $query
@@ -236,56 +276,63 @@ _apply_read_preferences_mongos (
        *
        * This applies to commands, too.
        */
-      result->query_with_read_prefs = bson_new ();
+      result->assembled_query = bson_new ();
       result->query_owned = true;
 
       if (bson_has_field (query_bson, "$query")) {
-         bson_concat (result->query_with_read_prefs, query_bson);
+         bson_concat (result->assembled_query, query_bson);
       } else {
          bson_append_document (
-            result->query_with_read_prefs, "$query", 6, query_bson);
+            result->assembled_query, "$query", 6, query_bson);
       }
 
       bson_append_document_begin (
-         result->query_with_read_prefs, "$readPreference", 15, &child);
+         result->assembled_query, "$readPreference", 15, &child);
       mode_str = _mongoc_read_mode_as_str (mode);
       bson_append_utf8 (&child, "mode", 4, mode_str, -1);
       if (!bson_empty0 (tags)) {
          bson_append_array (&child, "tags", 4, tags);
       }
 
-      max_staleness_seconds =
-         mongoc_read_prefs_get_max_staleness_seconds (read_prefs);
-
       if (max_staleness_seconds != MONGOC_NO_MAX_STALENESS) {
          bson_append_int64 (
             &child, "maxStalenessSeconds", 19, max_staleness_seconds);
       }
 
-      bson_append_document_end (result->query_with_read_prefs, &child);
+      if (!bson_empty0 (hedge)) {
+         bson_append_document (&child, "hedge", 5, hedge);
+      }
+
+      bson_append_document_end (result->assembled_query, &child);
    }
 }
 
 /*
  *--------------------------------------------------------------------------
  *
- * apply_read_preferences --
+ * assemble_query --
  *
- *       Update @result based on @read prefs, following the Server Selection
+ *       Update @result based on @read_prefs, following the Server Selection
  *       Spec.
  *
  * Side effects:
- *       Sets @result->query_with_read_prefs and @result->flags.
+ *       Sets @result->assembled_query and @result->flags.
+ *
+ *  Note:
+ *       This function, the mongoc_assemble_query_result_t struct, and all
+ *       related functions are only used for find operations with OP_QUERY.
+ *       Remove them once we have implemented exhaust cursors with OP_MSG in
+ *       the server, and all previous server versions are EOL.
  *
  *--------------------------------------------------------------------------
  */
 
 void
-apply_read_preferences (const mongoc_read_prefs_t *read_prefs,
-                        const mongoc_server_stream_t *server_stream,
-                        const bson_t *query_bson,
-                        mongoc_query_flags_t initial_flags,
-                        mongoc_apply_read_prefs_result_t *result /* OUT */)
+assemble_query (const mongoc_read_prefs_t *read_prefs,
+                const mongoc_server_stream_t *server_stream,
+                const bson_t *query_bson,
+                mongoc_query_flags_t initial_flags,
+                mongoc_assemble_query_result_t *result /* OUT */)
 {
    mongoc_server_description_type_t server_type;
 
@@ -296,7 +343,7 @@ apply_read_preferences (const mongoc_read_prefs_t *read_prefs,
    BSON_ASSERT (result);
 
    /* default values */
-   result->query_with_read_prefs = (bson_t *) query_bson;
+   result->assembled_query = (bson_t *) query_bson;
    result->query_owned = false;
    result->flags = initial_flags;
 
@@ -308,11 +355,11 @@ apply_read_preferences (const mongoc_read_prefs_t *read_prefs,
          _apply_read_preferences_mongos (read_prefs, query_bson, result);
       } else {
          /* Server Selection Spec: for topology type single and server types
-          * besides mongos, "clients MUST always set the slaveOK wire protocol
+          * besides mongos, "clients MUST always set the secondaryOk wire protocol
           * flag on reads to ensure that any server type can handle the
           * request."
           */
-         result->flags |= MONGOC_QUERY_SLAVE_OK;
+         result->flags |= MONGOC_QUERY_SECONDARY_OK;
       }
 
       break;
@@ -320,18 +367,19 @@ apply_read_preferences (const mongoc_read_prefs_t *read_prefs,
    case MONGOC_TOPOLOGY_RS_NO_PRIMARY:
    case MONGOC_TOPOLOGY_RS_WITH_PRIMARY:
       /* Server Selection Spec: for RS topology types, "For all read
-       * preferences modes except primary, clients MUST set the slaveOK wire
+       * preferences modes except primary, clients MUST set the secondaryOk wire
        * protocol flag to ensure that any suitable server can handle the
-       * request. Clients MUST  NOT set the slaveOK wire protocol flag if the
+       * request. Clients MUST  NOT set the secondaryOk wire protocol flag if the
        * read preference mode is primary.
        */
       if (read_prefs && read_prefs->mode != MONGOC_READ_PRIMARY) {
-         result->flags |= MONGOC_QUERY_SLAVE_OK;
+         result->flags |= MONGOC_QUERY_SECONDARY_OK;
       }
 
       break;
 
    case MONGOC_TOPOLOGY_SHARDED:
+   case MONGOC_TOPOLOGY_LOAD_BALANCED:
       _apply_read_preferences_mongos (read_prefs, query_bson, result);
       break;
 
@@ -347,14 +395,14 @@ apply_read_preferences (const mongoc_read_prefs_t *read_prefs,
 
 
 void
-apply_read_prefs_result_cleanup (mongoc_apply_read_prefs_result_t *result)
+assemble_query_result_cleanup (mongoc_assemble_query_result_t *result)
 {
    ENTRY;
 
    BSON_ASSERT (result);
 
    if (result->query_owned) {
-      bson_destroy (result->query_with_read_prefs);
+      bson_destroy (result->assembled_query);
    }
 
    EXIT;
